import { objectOmit } from '@dzeio/object-util'
import mongoose from 'mongoose'
import type APIKey from '.'
import Client from '../Client'
import Dao from '../Dao'

export default class APIKeyDao extends Dao<APIKey> {

	// @ts-expect-error typing fix
	private model = mongoose.models['APIKey'] as null ?? mongoose.model('APIKey', new mongoose.Schema({
		user: { type: String, required: true },
		key: { type: String, required: true, unique: true, index: true},
		permissions: [{ type: String }]
	}, {
		timestamps: true
	}))

	public async create(obj: Omit<APIKey, 'id' | 'created' | 'updated'>): Promise<APIKey | null> {
		await Client.get()
		return this.fromSource(await this.model.create(obj))
	}

	public async findAll(query?: Partial<APIKey> | undefined): Promise<APIKey[]> {
		await Client.get()
		try {
			if (query?.id) {
				const item = await this.model.findById(new mongoose.Types.ObjectId(query.id))
				if (!item) {
					return []
				}
				return [this.fromSource(item)]
			}
			const resp = await this.model.find(query ? this.toSource(query as APIKey) : {})
			return resp.map(this.fromSource)
		} catch (e) {
			console.error(e)
			return []
		}

	}

	public async update(obj: APIKey): Promise<APIKey | null> {
		await Client.get()

		const query = await this.model.updateOne({
			_id: new mongoose.Types.ObjectId(obj.id)
		}, this.toSource(obj))
		if (query.matchedCount >= 1) {
			obj.updated = new Date()
			return obj
		}
		return null
		// return this.fromSource()
	}

	public async delete(obj: APIKey): Promise<boolean> {
		await Client.get()
		const res = await this.model.deleteOne({
			_id: new mongoose.Types.ObjectId(obj.id)
		})
		return res.deletedCount > 0
	}

	private toSource(obj: APIKey): Omit<APIKey, 'id'> {
		return objectOmit(obj, 'id', 'updated', 'created')
	}

	private fromSource(doc: mongoose.Document<any, any, APIKey>): APIKey {
		return {
			id: doc._id.toString(),
			user: doc.get('user'),
			key: doc.get('key'),
			permissions: doc.get('permissions') ?? [],
			updated: doc.get('updatedAt'),
			created: doc.get('createdAt')
		}
	}

}
