import { objectOmit } from '@dzeio/object-util'
import mongoose from 'mongoose'
import type Config from '.'
import Client from '../Client'
import Dao from '../Dao'

export default class ConfigDao extends Dao<Config> {

	// @ts-expect-error typing fix
	private model = mongoose.models['Config'] as null ?? mongoose.model('Config', new mongoose.Schema({
		user: { type: String, required: true },
		type: { type: String, required: true},
		files: [{
			name: { type: String, unique: true, required: true},
			data: { type: Buffer, required: true }
		}]
	}, {
		timestamps: true
	}))


	public async create(obj: Omit<Config, 'id' | 'created' | 'updated'>): Promise<Config | null> {
		await Client.get()
		return this.fromSource(await this.model.create(obj))
	}

	public async findAll(query?: Partial<Config> | undefined): Promise<Config[]> {
		await Client.get()
		try {
			if (query?.id) {
				const item = await this.model.findById(new mongoose.Types.ObjectId(query.id))
				if (!item) {
					return []
				}
				return [this.fromSource(item)]
			}
			const resp = await this.model.find(query ? this.toSource(query as Config) : {})
			return resp.map(this.fromSource)
		} catch (e) {
			console.error(e)
			return []
		}

	}

	public async update(obj: Config): Promise<Config | null> {
		await Client.get()

		const query = await this.model.updateOne({
			_id: new mongoose.Types.ObjectId(obj.id)
		}, this.toSource(obj))
		if (query.matchedCount >= 1) {
			obj.updated = new Date()
			return obj
		}
		return null
		// return this.fromSource()
	}

	public async delete(obj: Config): Promise<boolean> {
		await Client.get()
		const res = await this.model.deleteOne({
			_id: new mongoose.Types.ObjectId(obj.id)
		})
		return res.deletedCount > 0
	}

	private toSource(obj: Config): Omit<Config, 'id'> {
		return objectOmit(obj, 'id', 'updated', 'created')
	}

	private fromSource(doc: mongoose.Document<any, any, Config>): Config {
		return {
			id: doc._id.toString(),
			user: doc.get('user'),
			type: doc.get('type'),
			files: doc.get('files') ?? [],
			updated: doc.get('updatedAt'),
			created: doc.get('createdAt')
		}
	}

}
